package ap.proof;

import ap.terfor.{Conjunction, Formula, TermOrder}
import ap.util.Debug

object AndTree {
  
  private val AC = Debug.AC_PROOF_TREE

  def apply(subtrees : Seq[ProofTree], order : TermOrder) : ProofTree =
    new AndTree(subtrees, order)
  
}

class AndTree private (val subtrees : Seq[ProofTree],
                       protected val order : TermOrder) extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(AndTree.AC, !subtrees.isEmpty)
  //////////////////////////////////////////////////////////////////////////////
  
  lazy val closingConstraint : Formula =
    Conjunction.conj(for (t <- subtrees.elements) yield t.closingConstraint,
                     order)


  //////////////////////////////////////////////////////////////////////////////

  private def indent(str : String, prefix : String) =
    prefix + str.replaceAll("\n", "\n" + prefix)
    
  override def toString : String = {
    "AndTree {\n" +
      (for (t <- subtrees) yield indent("" + t, "  ")).reduceLeft(
         (s1 : String, s2 : String) => s1 + "\n/\\\n" + s2) + 
    "\n}" + "\n" +
    "Constraint: " + closingConstraint
  }

}
