package ap.proof;

import ap.terfor.{Formula, TermOrder, Conjunction, Quantifier, NegatedConjunctions}
import ap.proof.goal.Goal
import ap.util.{Debug, Seqs}

object ConstraintSimplifier {
  
  private val AC = Debug.AC_CONSTRAINT_SIMPLIFIER  
  
  def apply(f : Formula, order : TermOrder) : Formula = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, order isSortingOf f)
    ////////////////////////////////////////////////////////////////////////////

    // TODO: check that the formula actually can be handled

    println("Simplify: " + f)
    
    firstQuantifier(f) match {
      case Some(Quantifier.ALL) =>
        Goal(f, order).runRunRun.closingConstraint
      case Some(Quantifier.EX) => {
        val goal = Goal(Conjunction.negate(f, order), order)
        Conjunction.negate(goal.runRunRun.closingConstraint, order)
      }
      case None => f // TODO
    }
  }
  
  /**
   * Find an outermost quantifier in <code>f</code> and return it, or return
   * <code>None</code> if no quantifier was found
   */
  private def firstQuantifier(f : Formula) : Option[Quantifier] = f match {
    case f : Conjunction =>
      if (f.isDivisibility || f.isNonDivisibility)
        None
      else if (f.quans.isEmpty)
        firstQuantifier(f.negatedConjs)
      else
        Some(f.quans.last)
    case f : NegatedConjunctions =>
      for (q <- Seqs.some(for (conj <- f.elements) yield firstQuantifier(conj)))
      yield q.dual
    case _ => None
  }
  
}
