package ap.proof;

import ap.terfor.{Quantifier, TermOrder, ConstantTerm, VariableTerm, Formula,
                  Conjunction}
import ap.terfor.substitutions.{Substitution, ConstantSubst, IdentitySubst,
                                ComposeSubsts}
import ap.util.{PlainRange, Debug}

object QuantifiedTree {
  
  private val AC = Debug.AC_PROOF_TREE
  
  def apply(subtree : ProofTree,
            quan : Quantifier,
            quantifiedConstants : Set[ConstantTerm],
            order : TermOrder,
            subtreeOrder : TermOrder) : ProofTree = subtree match {
    case subtree : QuantifiedTree if (subtree.quan == quan) =>
      new QuantifiedTree(subtree.subtree,
                         quan,
                         quantifiedConstants ++ subtree.quantifiedConstants,
                         order, subtree.subtreeOrder)
    case _ => 
      new QuantifiedTree(subtree, quan, quantifiedConstants, order, subtreeOrder)
  }
  
}

/**
 * <code>ProofTreeOneChild</code> that quantifies a set of constants in
 * the closing constraint of its <code>subtree</code>
 */
class QuantifiedTree private (val subtree : ProofTree,
                              val quan : Quantifier,
                              val quantifiedConstants : Set[ConstantTerm],
                              protected val order : TermOrder,
                              protected val subtreeOrder : TermOrder)
      extends ProofTreeOneChild {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(QuantifiedTree.AC,
                   !quantifiedConstants.isEmpty &&
                   (subtree match {
                     case subtree : QuantifiedTree => subtree.quan != this.quan
                     case _ => true
                    }))
  //////////////////////////////////////////////////////////////////////////////

  lazy val closingConstraint : Formula = {
    var constantSubst : Substitution = new IdentitySubst(subtreeOrder)
    for ((c, i) <- quantifiedConstants.elements.zipWithIndex) {
      val nextSubst = new ConstantSubst(c, VariableTerm(i), subtreeOrder)
      constantSubst = ComposeSubsts(constantSubst, nextSubst, subtreeOrder)
    }
    
    val quans : Seq[Quantifier] =
      (for (_ <- PlainRange(quantifiedConstants.size)) yield quan)    
    
    val constraintWithVars = constantSubst(subtree.closingConstraint)
    val nonSimplified =
      (Conjunction.quantify(quans, constraintWithVars, subtreeOrder) sortBy order)
      
    ConstraintSimplifier(nonSimplified, order)
  }
  
  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = {
    "QuantifiedTree (" + quan + ", " + quantifiedConstants + ") {\n" + subtree + "\n}"
  }

}
