package ap.proof.goal;

import ap.terfor.{Formula, Conjunction, Quantifier, ConstantTerm}
import ap.terfor.substitutions.VariableSubst
import ap.util.{Debug, PlainRange}

object QuantifierTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
}

abstract class QuantifierTask(conj : Conjunction, age : Int) extends Task {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(QuantifierTask.AC,
                   !conj.isTrue && !conj.isFalse && !conj.quans.isEmpty)
  //////////////////////////////////////////////////////////////////////////////
  
  val priority : Int = age

  /**
   * The name prefix to use for generated constants
   */
  protected def constantNameBase : String
  
  /**
   * Compute the new set of eliminated constants for a goal, given the set
   * of new constants (when carrying out this task)
   */
  protected def newEliminatedConstants(goal : Goal,
                                       newConstants : Iterable[ConstantTerm])
                                                            : Set[ConstantTerm]
    
                                                            
  /**
   * TODO: currently, gamma rules do not introduce a copy of the instantiated
   * formula
   */
  protected def instantiateWithConstants(goal : Goal,
                                         taskFinishedAction : TaskFinishedAction)
                                                               : ProofTree = {
    val (quan, num) = firstQuans
    val constants = for (i <- PlainRange(num))
                    yield new ConstantTerm (constantNameBase + goal.age + "_" + i)

    val newOrder = (goal.order /: constants)((o, c) => o.extend(c, Set.empty))
                                                                
    val instantiatedConjTask = {
      val instantiatedConj =
        new VariableSubst (0, constants, newOrder) (conj.unquantify(num))
      new ComplexFormulaTask (instantiatedConj)
    }

    val subtree = goal.taskFinished(newEliminatedConstants(goal, constants),
                                    newOrder,
                                    Array(instantiatedConjTask),
                                    taskFinishedAction)
    QuantifiedTree(subtree,
                   quan, Set.empty ++ constants,
                   goal.order, newOrder)
  }
                                                              
  /**
   * Determine the leading quantifier of the conjunction and the number of the
   * same quantifiers in a row
   *
   * TODO: do this in a smarter way (determine which quantifiers can be permuted)?   
   */
  private def firstQuans : (Quantifier, Int) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(QuantifierTask.AC, !conj.quans.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
                                                                
    val quans = conj.quans
                                                                  
    val quan = quans.last
    val num = quans.size - quans.lastIndexOf(quan.dual) - 1
    (quan, num)
  }                                                            

}
