package ap.terfor.substitutions;

import ap.util.{Debug, Logic}

object VariableShiftSubst {
  
  private val AC = Debug.AC_SUBSTITUTIONS
  
}

/**
 * Shift all variables with an index that is at least <code>offset</code>
 * upwards (<code>shift</code> steps). <code>shift</code> can be negative to
 * realise downwards shifting
 */
class VariableShiftSubst(offset : Int, shift : Int,
                         protected [substitutions] val order : TermOrder)
      extends SimpleSubstitution with Sorted[VariableShiftSubst] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(VariableShiftSubst.AC, offset >= 0 && offset + shift >= 0)
  //////////////////////////////////////////////////////////////////////////////     

  protected[substitutions] def passQuantifiers(num : Int) : Substitution =
    new VariableShiftSubst(offset + num, shift, order)
  
  protected[substitutions] def applyToVariable(v : VariableTerm) : Term = v match {
    case VariableTerm(i) => if (i < offset) v else VariableTerm(i + shift) 
  }

  protected[substitutions] def applyToConstant(c : ConstantTerm) : Term = c

  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean =
    Logic.forall(for (VariableTerm(i) <- t.variables.elements) yield i < offset)

  //////////////////////////////////////////////////////////////////////////////
    
  def sortBy(newOrder : TermOrder) : VariableShiftSubst =
    new VariableShiftSubst(offset, shift, newOrder)

}
